
function Adapter(obj) {
  if (!(this instanceof Adapter)) {
    return new Adapter(obj);
  }
  this.eventsList = [];
  var self = this;
  self.obj = obj;
}

/**
 * Default subscription method.
 * Subscribe to changes on the model.
 *
 * @param {Object} obj
 * @param {String} prop
 * @param {Function} fn
 */

Adapter.prototype.subscribe = function(prop, fn) {
  var model = this.obj;
  if (typeof model.on == 'function') {
    var name = 'change ' + prop;
    this.eventsList.push[{
      name: name,
      callback: fn
    }]
    model.on(name, fn);
  }
};

/**
 * Default unsubscription method.
 * Unsubscribe from changes on the model.
 */

Adapter.prototype.unsubscribe = function(prop, fn) {
  var model = this.obj;
  if (typeof model.off == 'function') {
    var name = 'change ' + prop;
    var eventsList = this.eventsList;
    //remove from eventsList
    for (var i = 0, len = eventsList.length; i < len; i++) {
      var o = eventsList[i];
      if (o.name == name && o.callback == fn) {
        eventsList.splice(i, 1);
        break;
      }
    }
    model.off(name, fn);
  }
};

/**
 * Remove all subscriptions on this adapter
 */

Adapter.prototype.unsubscribeAll = function() {
  var model = this.obj;
  if (typeof model.off == 'function') {
    this.eventsList.forEach(function(o, i) {
      model.off(o.name, o.callback);
    }, this);
    this.eventsList = [];
  }
};

/**
 * Default setter method.
 * Set a property on the model.
 *
 * @param {Object} obj
 * @param {String} prop
 * @param {Mixed} val
 */

Adapter.prototype.set = function(prop, val) {
  var obj = this.obj;
  if (obj == null) return;
  if ('function' == typeof obj[prop]) {
    obj[prop](val);
  }
  else if ('function' == typeof obj.set) {
    obj.set(prop, val);
  }
  else {
    obj[prop] = val;
  }
};

/**
 * Default getter method.
 * Get a property from the model.
 *
 * @param {Object} obj
 * @param {String} prop
 * @return {Mixed}
 */

Adapter.prototype.get = function(prop) {
  var obj = this.obj;
  if (obj == null) {
    return obj; // null or undefined
  }

  // split property on '.' access
  // and dig into the object
  var parts = prop.split('.');
  var part = parts.shift();
  do {
    if (typeof obj[part] === 'function') {
      obj = obj[part].call(obj);
    }
    else {
      obj = obj[part];
    }

    if (obj == null) {
      return obj; // null or undefined
    }

    part = parts.shift();
  } while(part);

  return obj;
};

module.exports = Adapter;
